/*************************************************** 
  www.buydisplay.com
 ****************************************************/

#include "TFTM1.65-2.h"
#include <limits.h>
#include "pins_arduino.h"
#include "wiring_private.h"
#include <SPI.h>
 #include "Arduino.h"

// Constructor when using software SPI.  All output pins are configurable.
NV3007::NV3007(int8_t cs, int8_t dc, int8_t mosi,
				   int8_t sclk, int8_t rst) : ERGFX(NV3007_TFTWIDTH, NV3007_TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _mosi  = mosi;
  _sclk = sclk;
  _rst  = rst;
  hwSPI = false;
}


// Constructor when using hardware SPI.  Faster, but must use SPI pins
// specific to each board type (e.g. 11,13 for Uno, 51,52 for Mega, etc.)
NV3007::NV3007(int8_t cs, int8_t dc, int8_t rst) : ERGFX(NV3007_TFTWIDTH, NV3007_TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _rst  = rst;
  hwSPI = true;
  _mosi  = _sclk = 0;
}

void NV3007::spiwrite(uint8_t c) {

  //Serial.print("0x"); Serial.print(c, HEX); Serial.print(", ");

  if (hwSPI) {
#if defined (__AVR__)
      uint8_t backupSPCR = SPCR;
    SPCR = mySPCR;
    SPDR = c;
    while(!(SPSR & _BV(SPIF)));
    SPCR = backupSPCR;
#elif defined(ESP32)|| defined(TEENSYDUINO)
    SPI.transfer(c);
#elif defined (__arm__)
    SPI.setClockDivider(11); // 8-ish MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    SPI.transfer(c);
#endif
  } else {
    // Fast SPI bitbang swiped from LPD8806 library
    for(uint8_t bit = 0x80; bit; bit >>= 1) {
      if(c & bit) {
	//digitalWrite(_mosi, HIGH); 
	*mosiport |=  mosipinmask;
      } else {
	//digitalWrite(_mosi, LOW); 
	*mosiport &= ~mosipinmask;
      }
      //digitalWrite(_sclk, LOW);
      *clkport &= ~clkpinmask;
       //digitalWrite(_sclk, HIGH);
      *clkport |=  clkpinmask;
    }
  }
}


void NV3007::writecommand(uint8_t c) {
  *dcport &=  ~dcpinmask;
  //digitalWrite(_dc, LOW);
  //*clkport &= ~clkpinmask; // clkport is a NULL pointer when hwSPI==true
  //digitalWrite(_sclk, LOW);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  spiwrite(c);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
}


void NV3007::writedata(uint8_t c) {
  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  //*clkport &= ~clkpinmask; // clkport is a NULL pointer when hwSPI==true
  //digitalWrite(_sclk, LOW);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);
  
  spiwrite(c);

  //digitalWrite(_cs, HIGH);
  *csport |= cspinmask;
} 
/*
// If the SPI library has transaction support, these functions
// establish settings and protect from interference from other
// libraries.  Otherwise, they simply do nothing.
#ifdef SPI_HAS_TRANSACTION
static inline void spi_begin(void) __attribute__((always_inline));
static inline void spi_begin(void) {
  SPI.beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0));
}
static inline void spi_end(void) __attribute__((always_inline));
static inline void spi_end(void) {
  SPI.endTransaction();
}
#else
#define spi_begin()
#define spi_end()
#endif
*/

void NV3007::begin(void) {
  if (_rst > 0) {
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, LOW);
  }

  pinMode(_dc, OUTPUT);
  pinMode(_cs, OUTPUT);
  
  csport    = portOutputRegister(digitalPinToPort(_cs));
  cspinmask = digitalPinToBitMask(_cs);
  dcport    = portOutputRegister(digitalPinToPort(_dc));
  dcpinmask = digitalPinToBitMask(_dc);

  if(hwSPI) { // Using hardware SPI
#if defined (__AVR__)
    SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    mySPCR = SPCR;
 #elif defined(ESP32)
    SPI.begin();
  SPI.beginTransaction(SPISettings(40000000, MSBFIRST, SPI_MODE0));  
   // SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
   // SPI.setBitOrder(MSBFIRST);
   // SPI.setDataMode(SPI_MODE0);
   
#elif defined(TEENSYDUINO)
    SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
#elif defined (__arm__)
      SPI.begin();
      SPI.setClockDivider(11); // 8-ish MHz (full! speed!)
      SPI.setBitOrder(MSBFIRST);
      SPI.setDataMode(SPI_MODE0);
#endif
  } else {
    pinMode(_sclk, OUTPUT);
    pinMode(_mosi, OUTPUT);

    clkport     = portOutputRegister(digitalPinToPort(_sclk));
    clkpinmask  = digitalPinToBitMask(_sclk);
    mosiport    = portOutputRegister(digitalPinToPort(_mosi));
    mosipinmask = digitalPinToBitMask(_mosi);
    *clkport   &= ~clkpinmask;
    *mosiport  &= ~mosipinmask;
  }

  // toggle RST low to reset
  if (_rst > 0) {
    digitalWrite(_rst, HIGH);
    delay(5);
    digitalWrite(_rst, LOW);
    delay(20);
    digitalWrite(_rst, HIGH);
    delay(150);
  }

  if (hwSPI) SPI.begin();
  
 
 	writecommand(0xff);
	writedata(0xa5);
	writecommand(0x9a);
	writedata(0x08);
	writecommand(0x9b);
	writedata(0x08);
	writecommand(0x9c);
	writedata(0xb0);
	writecommand(0x9d);
	writedata(0x17);
	writecommand(0x9e);
	writedata(0xc2);
	writecommand(0x8f);
	writedata(0x22);
	writedata(0x04);
	writecommand(0x84);
	writedata(0x90);
	writecommand(0x83);
	writedata(0x7B);
	writecommand(0x85);
	writedata(0x4F);
    ////GAMMA---------------------------------/////////////
    //V0[3:0]
	writecommand(0x6e);
	writedata(0x0f);
	writecommand(0x7e);
	writedata(0x0f);

    //V63[3:0]
	writecommand(0x60);
	writedata(0x00);
	writecommand(0x70);
	writedata(0x00);
    //V1[5:0] 
	writecommand(0x6d);
	writedata(0x39);
	writecommand(0x7d);
	writedata(0x31);
    //V62[5:0]
	writecommand(0x61);
	writedata(0x0A);
	writecommand(0x71);
	writedata(0x0A);
    //V2[5:0]
	writecommand(0x6c);
	writedata(0x35);
	writecommand(0x7c);
	writedata(0x29);
    //V61[5:0]
	writecommand(0x62);
	writedata(0x0F);
	writecommand(0x72);
	writedata(0x0F);
    //V20[6:0]
	writecommand(0x68);
	writedata(0x4f);
	writecommand(0x78);
	writedata(0x45);
    //V43[6:0]
	writecommand(0x66);
	writedata(0x33);
	writecommand(0x76);
	writedata(0x33);
    //V4[4:0]
	writecommand(0x6b);
	writedata(0x14);
	writecommand(0x7b);
	writedata(0x14);
    //V59[4:0]
	writecommand(0x63);
	writedata(0x09);
	writecommand(0x73);
	writedata(0x09);
    //V6[4:0]
	writecommand(0x6a);
	writedata(0x13);
	writecommand(0x7a);
	writedata(0x16);
    //V57[4:0]
	writecommand(0x64);
	writedata(0x08);
	writecommand(0x74);
	writedata(0x08);
	writecommand(0x69);
	writedata(0x07);
	writecommand(0x79);
	writedata(0x0d);
	writecommand(0x65);
	writedata(0x05);
	writecommand(0x75);
	writedata(0x05);
	writecommand(0x67);
	writedata(0x33);
	writecommand(0x77);
	writedata(0x33);
	writecommand(0x6f);
	writedata(0x00);
	writecommand(0x7f);
	writedata(0x00);
	writecommand(0x50);
	writedata(0x00);
	writecommand(0x52);
	writedata(0xd6);
	writecommand(0x53);
	writedata(0x04);
	writecommand(0x54);
	writedata(0x04);
	writecommand(0x55);
	writedata(0x1b);
	writecommand(0x56);
	writedata(0x1b);
	writecommand(0xa0);
	writedata(0x2a);
	writedata(0x24);
	writedata(0x00);
	writecommand(0xa1);
	writedata(0x84);
	writecommand(0xa2);
	writedata(0x85);
	writecommand(0xa8);
	writedata(0x34);
	writecommand(0xa9);
	writedata(0x80);
	writecommand(0xaa);
	writedata(0x73);
	writecommand(0xAB);
	writedata(0x03);
	writedata(0x61);
	writecommand(0xAC);
	writedata(0x03);
	writedata(0x65);
	writecommand(0xAD);
	writedata(0x03);
	writedata(0x60);
	writecommand(0xAE);
	writedata(0x03);
	writedata(0x64);
	writecommand(0xB9);
	writedata(0x82);
	writecommand(0xBA);
	writedata(0x83);
	writecommand(0xBB);
	writedata(0x80);
	writecommand(0xBC);
	writedata(0x81);
	writecommand(0xBD);
	writedata(0x02);
	writecommand(0xBE);
	writedata(0x01);
	writecommand(0xBF);
	writedata(0x04);
	writecommand(0xC0);
	writedata(0x03);
	writecommand(0xc4);
	writedata(0x33);
	writecommand(0xc5);
	writedata(0x80);
	writecommand(0xc6);
	writedata(0x73);
	writecommand(0xc7);
	writedata(0x00);
	writecommand(0xC8);
	writedata(0x33);
	writedata(0x33);
	writecommand(0xC9);
	writedata(0x5b);
	writecommand(0xCA);
	writedata(0x5a);
	writecommand(0xCB);
	writedata(0x5d);
	writecommand(0xCC);
	writedata(0x5c);
	writecommand(0xCD);
	writedata(0x33);
	writedata(0x33);
	writecommand(0xCE);
	writedata(0x5f);
	writecommand(0xCF);
	writedata(0x5e);
	writecommand(0xD0);
	writedata(0x61);
	writecommand(0xD1);
	writedata(0x60);
	writecommand(0xB0);
	writedata(0x3a);
	writedata(0x3a);
	writedata(0x00);
	writedata(0x00);
	writecommand(0xB6);
	writedata(0x32);
	writecommand(0xB7);
	writedata(0x80);
	writecommand(0xB8);
	writedata(0x73);
	writecommand(0xe0);
	writedata(0x00);
	writecommand(0xe1);
	writedata(0x03);
	writedata(0x0f);
	writecommand(0xe2);
	writedata(0x04);
	writecommand(0xe3);
	writedata(0x01);
	writecommand(0xe4);
	writedata(0x0e);
	writecommand(0xe5);
	writedata(0x01);
	writecommand(0xe6);
	writedata(0x19);
	writecommand(0xe7);
	writedata(0x10);
	writecommand(0xe8);
	writedata(0x10);
	writecommand(0xe9);
	writedata(0x21);
	writecommand(0xea);
	writedata(0x12);
	writecommand(0xeb);
	writedata(0xd0);
	writecommand(0xec);
	writedata(0x04);
	writecommand(0xed);
	writedata(0x07);
	writecommand(0xee);
	writedata(0x07);
	writecommand(0xef);
	writedata(0x09);
	writecommand(0xF0);
	writedata(0xD0);
	writecommand(0xF1);
	writedata(0x0E);
	writecommand(0xF9);
	writedata(0x56);
	writecommand(0xf2);
	writedata(0x26);
	writedata(0x1b);
	writedata(0x0b);
	writedata(0x20);
	writecommand(0xec);
	writedata(0x04);
	writecommand(0x35);
	writedata(0x00);
	writecommand(0x44);
	writedata(0x00);
	writedata(0x10);
	writecommand(0x46);
	writedata(0x10);
	writecommand(0xff);
	writedata(0x00);
	writecommand(0x3a);
	writedata(0x05);
	writecommand(0x36);
    	writedata(0x00);


        writecommand(0x11);    //Exit Sleep 
      //  if (hwSPI) spi_end();
        delay(120); 		
       // if (hwSPI) spi_begin();
        writecommand(0x29);    //Display on 
       // if (hwSPI) spi_end();

}


void NV3007::setAddrWindow(uint16_t x0, uint16_t y0, uint16_t x1,
 uint16_t y1) {
  x0+=0x0c;x1+=0x0c;
  writecommand(NV3007_CASET); // Column addr set
  writedata(x0 >> 8);
  writedata(x0 & 0xFF);     // XSTART 
  writedata(x1 >> 8);
  writedata(x1 & 0xFF);     // XEND

  writecommand(NV3007_PASET); // Row addr set
  writedata(y0>>8);
  writedata(y0);     // YSTART
  writedata(y1>>8);
  writedata(y1);     // YEND

  writecommand(NV3007_RAMWR); // write to RAM
  
  
}


void NV3007::pushColor(uint16_t color) {
 // if (hwSPI) spi_begin();
  //digitalWrite(_dc, HIGH);
  *dcport |=  dcpinmask;
  //digitalWrite(_cs, LOW);
  *csport &= ~cspinmask;

  spiwrite(color >> 8);
  spiwrite(color);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
 // if (hwSPI) spi_end();
}

void NV3007::drawPixel(int16_t x, int16_t y, uint16_t color) {

  if((x < 0) ||(x >= _width) || (y < 0) || (y >= _height)) return;

 // if (hwSPI) spi_begin();
  setAddrWindow(x,y,x+1,y+1);

  //digitalWrite(_dc, HIGH);
  *dcport |=  dcpinmask;
  //digitalWrite(_cs, LOW);
  *csport &= ~cspinmask;

  spiwrite(color >> 8);
  spiwrite(color);

  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
 // if (hwSPI) spi_end();
}


void NV3007::drawFastVLine(int16_t x, int16_t y, int16_t h,
 uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;

  if((y+h-1) >= _height) 
    h = _height-y;

 // if (hwSPI) spi_begin();
  setAddrWindow(x, y, x, y+h-1);

  uint8_t hi = color >> 8, lo = color;

  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  while (h--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
  //if (hwSPI) spi_end();
}


void NV3007::drawFastHLine(int16_t x, int16_t y, int16_t w,
  uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;
  if((x+w-1) >= _width)  w = _width-x;
//  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y);

  uint8_t hi = color >> 8, lo = color;
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
  //digitalWrite(_dc, HIGH);
  //digitalWrite(_cs, LOW);
  while (w--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  *csport |= cspinmask;
  //digitalWrite(_cs, HIGH);
 // if (hwSPI) spi_end();
}

void NV3007::fillScreen(uint16_t color) {
  fillRect(0, 0,  _width, _height, color);
}

// fill a rectangle
void NV3007::fillRect(int16_t x, int16_t y, int16_t w, int16_t h,
  uint16_t color) {

  // rudimentary clipping (drawChar w/big text requires this)
  if((x >= _width) || (y >= _height)) return;
  if((x + w - 1) >= _width)  w = _width  - x;
  if((y + h - 1) >= _height) h = _height - y;

  //if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y+h-1);

  uint8_t hi = color >> 8, lo = color;

  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

  for(y=h; y>0; y--) {
    for(x=w; x>0; x--) {
      spiwrite(hi);
      spiwrite(lo);
    }
  }
  //digitalWrite(_cs, HIGH);
  *csport |= cspinmask;
 // if (hwSPI) spi_end();
}


void NV3007::bmp16(uint16_t x, uint16_t y, const uint8_t *pBmp, uint16_t chWidth, uint16_t chHeight) {
    uint16_t i, j;
    uint16_t hwColor;
     uint8_t hi , lo ;

  // rudimentary clipping (drawChar w/big text requires this)
  if((x >= _width) || (y >= _height)) return;
  if((x + chWidth - 1) >= _width)  chWidth = _width  - x;
  if((y + chHeight - 1) >= _height) chHeight = _height - y;

 // if (hwSPI) SPI.begin();
    setAddrWindow(x, y, x+chWidth-1, y+chHeight-1);

  *dcport |=  dcpinmask;
  //digitalWrite(_dc, HIGH);
  *csport &= ~cspinmask;
  //digitalWrite(_cs, LOW);

    for(j = 0; j < chHeight; j++) {
        for(i = 0; i < chWidth; i++) {
          hi= pgm_read_byte(pBmp++);lo= pgm_read_byte(pBmp++);
          spiwrite(hi);
          spiwrite(lo);
        }
    }
   //digitalWrite(_cs, HIGH);
  *csport |= cspinmask;
 // if (hwSPI) SPI.end();   
}






// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t NV3007::color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

 

